freeslot('sfx_relina', 'sfx_silent', 'SPR2_BIGT')
freeslot('S_PLAY_RELINA_BIGT_START', 'S_PLAY_RELINA_BIGT', 'S_PLAY_RELINA_BIGT_STOP', 'S_PLAY_RELINA_STND_OR_SPINDASH')

sfxinfo[sfx_relina].caption = 'Spinning menace'
sfxinfo[sfx_silent].caption = '/' -- no caption

local relina_pull_everything = false

local always_pullable_objtypes = {}
for i,v in ipairs({MT_EGGGUARD, MT_FLINGRING, MT_FLINGCOIN, MT_FLINGBLUESPHERE, MT_FLINGEMERALD,
  MT_FLINGNIGHTSCHIP, MT_FLINGNIGHTSSTAR, MT_REDFLAG, MT_BLUEFLAG, MT_BOXSPARKLE,
  MT_SMALLBUBBLE, MT_MEDIUMBUBBLE, MT_EXTRALARGEBUBBLE, MT_FIREFLOWER}) do
	always_pullable_objtypes[v] = true
end
for i=0,6 do always_pullable_objtypes[MT_EMERALD1+i] = true end
for i=0,17 do always_pullable_objtypes[MT_FLICKY_01+i+i] = true end
addHook('MapLoad', function()
	if rawget(_G, 'MT_COOKIE') ~= nil then
		always_pullable_objtypes[MT_COOKIE] = true
	end
end)

addHook('NetVars', function(network)
	relina_pull_everything = network(relina_pull_everything)
end)

local function MobjPullable(relina, mobj)
	if mobj.relina_pullable then
		return mobj
	elseif mobj.relina_pullable == false then
		return nil
	end
	
	if mobj.flags & (MF_PUSHABLE|MF_ENEMY|MF_BOXICON|MF_MISSILE) then
		return mobj
	end
	
	if mobj.flags & MF_MONITOR then
		if mobj.flags & MF_GRENADEBOUNCE then
			-- Golden monitor, so spawn its icon and use that.
			if mobj.flags2 & MF2_AMBUSH then
				P_KillMobj(mobj, relina, relina)
			end
			local icon = P_SpawnMobjFromMobj(mobj, 0, 0, 0, mobjinfo[mobj.type].damage)
			icon.target = relina
			return icon
		else
			if mobj.state ~= S_BOX_POP1 and mobj.state ~= S_BOX_POP2 then
				if P_AproxDistance(relina.x-mobj.x, relina.y-mobj.y) < 64*relina.scale then
					P_DamageMobj(mobj, relina, relina, 1)
				end
			end
			return mobj
		end
	end
	
	local static_replace_types = {[MT_RING]=MT_FLINGRING, [MT_COIN]=MT_FLINGCOIN, [MT_BLUESPHERE]=MT_FLINGBLUESPHERE,
		[MT_NIGHTSCHIP]=MT_FLINGNIGHTSCHIP, [MT_NIGHTSSTAR]=MT_FLINGNIGHTSSTAR}
	if static_replace_types[mobj.type] then
		local sphere = P_SpawnMobjFromMobj(mobj, 0, 0, 0, static_replace_types[mobj.type])
		mobj.state = S_NULL
		return sphere
	elseif mobj.type == MT_PLAYER then
		if not (mobj.player and mobj.player.valid and mobj.player.relina_pull_disabled) then
			return mobj
		end
	elseif relina_pull_everything or always_pullable_objtypes[mobj.type] then
		return mobj
	end
end

local function RelinaPullMobj(relina, mobj, thrust)
	local pullable = MobjPullable(relina, mobj)
	if pullable == nil then return nil end
	local angle = R_PointToAngle2(pullable.x, pullable.y, relina.x, relina.y)
	P_Thrust(pullable, angle, thrust)
	if abs(pullable.z - relina.z) > 4*relina.scale then
		if pullable.z > relina.z then
			pullable.momz = $-thrust
		else
			pullable.momz = $+thrust
		end
	end
	return pullable
end

local function DetermineUpwardThrustMultiplier(mobj)
	if mobj.type == MT_PLAYER or (mobj.flags & MF_MISSILE) then
		return 10
	elseif not (mobj.flags & MF_SPECIAL|MF_PUSHABLE) then
		return 7
	else
		return 0
	end
end

function A_RelinaBigTornado(actor, var1, var2)
	if actor.frame & FF_FRAMEMASK > var2 or var2 == -1 then
		actor.frame = ($ & ~FF_FRAMEMASK) | var1
	end
	
	if actor.player and actor.player.valid then
		actor.player.dashspeed = 0
		if actor.player.cmd then
			local forward = actor.player.cmd.forwardmove
			local side = actor.player.cmd.sidemove
			if forward ~= 0 or side ~= 0 then
				local movementAngle = R_PointToAngle2(0, 0, forward*FRACUNIT, -side*FRACUNIT) + actor.angle
				local movespeed = 4*actor.scale
				actor.momx = FixedMul(cos(movementAngle), movespeed)
				actor.momy = FixedMul(sin(movementAngle), movespeed)
			end
		end
	end
	
	local range = 512*actor.scale
	local sightcheck_min = 512*FRACUNIT
	local closerange = 64*actor.scale
	local thrust = 4*actor.scale
	searchBlockmap('objects', function(refmobj, foundmobj)
		local dist = P_AproxDistance(P_AproxDistance(refmobj.x-foundmobj.x, refmobj.y-foundmobj.y), refmobj.z-foundmobj.z)
		if dist <= range and (dist < sightcheck_min or P_CheckSight(refmobj, foundmobj)) then
			local pullable = RelinaPullMobj(refmobj, foundmobj, thrust)
			if pullable ~= nil and dist <= closerange and pullable.valid then
				local upthrust = DetermineUpwardThrustMultiplier(pullable)*thrust
				if upthrust > 0 then
					if pullable.flags2 & MF2_OBJECTFLIP then
						upthrust = -$
					end
					pullable.momz = $+upthrust
				end
			end
		end
	end, actor, actor.x-range, actor.x+range, actor.y-range, actor.y+range)
end

function A_RelinaBigtStop(actor, var1, var2)
	if actor.player and actor.player.valid then
		if actor.player.pflags & PF_STARTDASH then
			actor.state = S_PLAY_SPINDASH
		elseif actor.player.pflags & PF_SPINNING then
			actor.state = S_PLAY_ROLL
		end
	end
end

states[S_PLAY_RELINA_BIGT_START] = {SPR_PLAY, SPR2_BIGT, 8, nil, 0, 0, S_PLAY_RELINA_BIGT}
states[S_PLAY_RELINA_BIGT] = {SPR_PLAY, SPR2_BIGT, 2, A_RelinaBigTornado, E, H, S_PLAY_RELINA_BIGT}
states[S_PLAY_RELINA_BIGT_STOP] = {SPR_PLAY, SPR2_BIGT, 8, A_RelinaBigTornado, I, -1, S_PLAY_RELINA_STND_OR_SPINDASH}
states[S_PLAY_RELINA_STND_OR_SPINDASH] = {SPR_PLAY, SPR2_BIGT, 0, A_RelinaBigtStop, 0, 0, S_PLAY_STND}

addHook('PlayerThink', function(player)
	if player.valid and player.mo and player.mo.valid and player.mo.skin == 'relina' then
		local sprites_with_sound = {[SPR2_SPIN]=true, [SPR2_BIGT]=true, [SPR2_ROLL]=true, [SPR2_JUMP]=true, [SPR2_NDRL]=true}
		local sound_should_play = sprites_with_sound[player.mo.sprite2] or false
		if sound_should_play ~= S_SoundPlaying(player.mo, sfx_relina) then
			if sound_should_play then
				S_StartSound(player.mo, sfx_relina)
			else
				S_StopSoundByID(player.mo, sfx_relina)
			end
		end
		
		if player.keep_in_bigt == nil then
			player.keep_in_bigt = false
		end
		
		if player.mo.state == S_PLAY_RELINA_BIGT or player.keep_in_bigt then
			if player.cmd.buttons & (BT_SPIN|BT_JUMP) == BT_SPIN|BT_JUMP then
				player.mo.state = S_PLAY_RELINA_BIGT
			else
				player.mo.state = S_PLAY_RELINA_BIGT_STOP
				player.keep_in_bigt = false
			end
		elseif player.mo.state == S_PLAY_RELINA_BIGT_STOP and (leveltime & 1) then
			player.mo.frame = $+1
			if player.mo.frame > L then
				player.mo.frame = I
			end
		end
		
		if not (player.cmd.buttons & BT_SPIN) then
			player.relina_bounce_held = false
		end
		
		if player.relina_bounce == nil then
			player.relina_bounce = false
			player.relina_bounce_held = false
		elseif player.relina_bounce and P_IsObjectOnGround(player.mo) then
			player.relina_bounce = false
			local shield = player.powers[pw_shield]
			if shield == SH_ELEMENTAL then
				P_ElementalFire(player, true)
			end
			P_DoBubbleBounce(player)
			player.mo.state = S_PLAY_JUMP
		end
		
		if player.mo.eflags & MFE_SPRUNG then
			player.relina_bounce = false
		end
	end
end)

addHook('JumpSpecial', function(player)
	if player.valid and player.mo and player.mo.valid and player.mo.skin == 'relina' then
		if player.mo.state == S_PLAY_SPINDASH then
			player.mo.state = S_PLAY_RELINA_BIGT_START
			player.keep_in_bigt = true
			return true
		elseif player.mo.state == S_PLAY_RELINA_BIGT_STOP then
			return true
		end
	end
	return false
end)

addHook('JumpSpinSpecial', function(player)
	if player.valid and player.mo and player.mo.valid and player.mo.skin == 'relina' then
		if not player.relina_bounce and not player.relina_bounce_held then
			local shield = player.powers[pw_shield]
			if (shield & (SH_PROTECTELECTRIC|SH_WHIRLWIND)) then
				if player.powers[pw_shield] ~= SH_ATTRACT or (player.mo.target and player.mo.target.valid) then
					return false
				end
			end
			player.mo.state = S_PLAY_JUMP
			S_StartSound(player.mo, sfx_s3k44)
			P_SetObjectMomZ(player.mo, -32*FRACUNIT)
			player.relina_bounce = true
			player.relina_bounce_held = true
			return true
		end
	end
end)

addHook('SpinSpecial', function(player)
	if player.valid and player.mo and player.mo.valid and player.mo.skin == 'relina' then
		if player.pflags & PF_SPINNING and not player.relina_bounce then
			if player.mo.state == S_PLAY_ROLL then
				player.pflags = $ | PF_STARTDASH
				player.mo.state = S_PLAY_SPINDASH
				player.dashspeed = max(16*FRACUNIT, P_AproxDistance(player.mo.momx, player.mo.momz))
				local angle = R_PointToAngle2(0, 0, player.mo.momx, player.mo.momy)
				local speed = min(P_AproxDistance(player.mo.momx, player.mo.momy), 4*FRACUNIT)
				player.mo.momx = FixedMul(cos(angle), speed)
				player.mo.momy = FixedMul(sin(angle), speed)
				return true
			elseif player.mo.state == S_PLAY_SPINDASH then
				player.dashspeed = $ + 3*FRACUNIT/2
				return true
			end
		end
	end
end)

addHook('ShieldSpecial', function(player)
	if player.valid and player.mo and player.mo.valid and player.mo.skin == 'relina' then
		local shield = player.powers[pw_shield]
		if (shield & SH_FORCE) or shield == SH_ELEMENTAL then
			return true
		elseif shield == SH_ATTRACT then
			local enemy = P_LookForEnemies(player)
			if not (enemy and enemy.valid and P_CheckSight(player.mo, enemy)) then
				return true
			end
		end
	end
end)	

addHook('ShouldDamage', function(target, inflictor, source, damage, damagetype)
	if target.valid and inflictor then
		if (inflictor.flags & (MF_ENEMY|MF_MISSILE|MF_PUSHABLE|MF_MONITOR)) and target.state == S_PLAY_RELINA_BIGT then
			return false
		end
	end
end)

local function ToggleCmdFn(setter_fn, status_fn)
	return function(player, arg)
		local a = (arg or ''):lower()
		if a == '0' or a == 'n' or a == 'no' or a == 'off' or a == 'false' or a == 'f' then
			setter_fn(player, false)
		elseif a == '1' or a == 'y' or a == 'yes' or a == 'on' or a == 'true' or a == 't' then
			setter_fn(player, true)
		end
		if consoleplayer == player then
			status_fn(player)
		end
	end
end

COM_AddCommand('relinapull', ToggleCmdFn(function(player, value)
	player.relina_pull_disabled = not value
end, function(player)
	if player.relina_pull_disabled then
		print("You will not be affected by Relina's tornado form.")
		print('To allow her to pull you, run "relinapull 1".')
	else
		print("You may be pulled around by Relina's tornado form.")
		print('To disable this, run "relinapull 0".')
	end
end))

COM_AddCommand('relinapulleverything', ToggleCmdFn(function(player, value)
	relina_pull_everything = value
end, function(player)
	if relina_pull_everything then
		print("Relina's tornado form can pull everything! (except for players with relinapull off)")
		print('Warning: this may break stuff.')
		print('To go back to the normal behavior, run "relinapulleverything 0".')
	else
		print("Relina's tornado form will only pull selected objects.")
		print('Run "relinapulleverything 1" if you dare.')
	end
end), COM_ADMIN)

hud.add(function(v, stplyr)
	if stplyr.mo and stplyr.mo.valid and stplyr.mo.skin == 'relina' and stplyr.mo.state == S_PLAY_SPINDASH then
		v.drawString(3, 193, 'SPIN POWER: '..tostring(stplyr.dashspeed/FRACUNIT), V_SNAPTOBOTTOM|V_PERPLAYER|V_MAGENTAMAP, 'small')
	end
end, 'game')